@@ -1,6 +1,6 @@
 #!/bin/sh
 #  Alicq ICQ client
-#  Copyright (C) Ihar Viarheichyk 2001
+#  Copyright (C) Ihar Viarheichyk 2001-2004
 
 #  This program is free software; you can redistribute it and/or modify
 #  it under the terms of the GNU General Public License as published by
@@ -15,12 +15,16 @@
 # Restart with wish \
 exec wish "$0" "$@"
 
-set base [file dirname [info script]]
+set base [file join [pwd] [file dirname [info script]]]
 lappend auto_path $base [file join $base .. share alicq lib]
 
+if {[catch { package require Tcl 8.3 }]} {
+	puts "Tcl 8.3 or higher is required to run Alicq"
+	exit -1
+}
 package require alicq 0.8.8
-package require AlicqWidgets
+package require AlicqWidgets 0.3
 
 puts [time alicq_init 1]
 
-# vim:syn=tcl
+# vim: filetype=tcl
