@@ -10,14 +10,16 @@ namespace eval meta {
 	array set words {
 	    type text save change
 	    default {sex porn fuck teen hot {\$[0-9]+} XXX hardcore}
-	    description "Words and regular expression which can indicate spam messages"
+	    description "Words or regular\nexpression which can\nindicate spam messages"
 	}
 	array set reply {
 		type text save change
 	        default "You are not in my contact list and sent me message which was considered as spam and rejected."
-	        description "Reply which is sent to sender of 'spam' message"
+	        description "Reply which is\nsent to sender of\n'spam' message"
+	}
+	array set log {type file save change 
+		description "File to save spam\nmessages to"
 	}
-	array set log {}
 }
 
 handler Incoming BlockSpam {type uid time message args} {
@@ -33,7 +35,7 @@ handler Incoming BlockSpam {type uid time message args} {
 				puts $f "From $uid at [clock format $time -format {%x %X}]\n$message\n---------------------"
 				close $f
 			}
-			Event Log 8 "Spam blocked from $uid: $message"
+			Event Log info "Spam blocked from $uid: $message"
 			return -code break
 		}
 	}
