@@ -6,7 +6,7 @@ package require Tk
 package require BWidget
 
 namespace eval meta {
-	set author "Ihat Viarheickyk <iverg@mail.ru>"
+	set author "Ihar Viarheichyk <iverg@mail.ru>"
 	set description "Basic implementation of chat dialogs."
 	array set popup {
 		type boolean default 0 property Global:UI:Messages|Popup
@@ -17,21 +17,21 @@ namespace eval meta {
 		type boolean default 0 save exit menu {Interface Tabbing}
 		description "Use tabbed panels within single wnidow"
 	}
-	array set add:contact {type action menu {Add Contact}}
-	array set add:group {type action menu {Add Group}}
 }
 
-option add *timeFormat {%X %x} widgetDefault
+option add *timeFormat %X widgetDefault
+option add *dateFormat %x widgetDefault
 option add *Text.underlineTag alias widgetDefault
-option add *Text.headerBackground gray85
+option add *Text.headerBackground gray85 widgetDefault
+option add *Text.bodyColor black widgetDefault
 
 option add *EditBar.send.text [mc Send] widgetDefault
-option add *EditBar.close.text x widgetDefault
+option add *EditBar.send.default active widgetDefault
 option add *EditBar.close.relief flat widgetDefault
 option add *EditBar.type.direction above widgetDefault
 option add *EditBar.type.anchor w widgetDefault
 option add *EditBar.AuxBar.cite.text [mc Cite] widgetDefault
-option add *Tablist.Radiobutton.selectColor Green widgetDefault
+option add *Tablist.Radiobutton.selectColor gray95 widgetDefault
 option add *Tablist.Radiobutton.foreground black widgetDefault
 option add *Tablist.Radiobutton.highlight blue widgetDefault
 
@@ -44,11 +44,12 @@ foreach {item value} {incoming	darkMagenta	outgoing	red
 
 event add <<Send>> <Control-Return>
 event add <<Cite>> <Control-q>
+event add <<History>> <Control-h>
 event add <<NextTab>> <Control-Right>
 event add <<PrevTab>> <Control-Left>
 foreach x {1 2 3 4 5 6 7 8 9} { event add <<ToTab>> <Control-Key-$x> }
 
-proc WinByUid {prefix uid} { return .${prefix}[string map {: -} $uid] }
+proc WinByUid {prefix uid} { return .${prefix}[string map {: - . -} $uid] }
 
 proc members {uid} {
 	if {[string match Contact:* $uid]} {
@@ -59,6 +60,11 @@ proc members {uid} {
 	}
 }
 
+# Create toplevel with saveable geometry
+proc Toplevel {name class} {
+	toplevel $name -class $class
+	bindtags $name [concat [bindtags $name] SaveGeometry]
+}
 # Raise windows, but not take focus if another dialog window has it
 proc Raise {win {flag 0}} {
 	set focus [focus]
@@ -73,20 +79,29 @@ proc Raise {win {flag 0}} {
 proc SingleDialog {uid {flag 0}} {
 	set top [WinByUid msg $uid]
 	if {![winfo exists $top]} {
-		toplevel $top -class AlicqUserWindow
-		wm title $top "[get $uid Alias] ([lrange [split $uid :] 1 2])"
+		Toplevel $top AlicqUserWindow
+		ChatTitle $top $uid "Alicq"
 		DialogFrame $top $uid
 	}
 	Raise $top $flag
 }
 
+# Update chat window title
+# TODO: invoke on status change
+proc ChatTitle {top uid text} {
+	set alias [get $uid Alias ""]
+	if {$alias==""} { set alias [lindex [split $uid :] end] }
+	set id [lrange [split $uid :] end-1 end]
+	set status [string tolower [mc [get $uid Status offline]]]
+	wm title $top "$alias ($status, $id) - $text"
+}
+
 # Main dialog for text messages used in tabbed window mode
 proc TabbedDialog {uid {flag 0}} {
 	set top .all
 	set var [namespace current]::frameSelected
 	if {![winfo exists $top]} {
-		toplevel $top -class AlicqUserWindow
-		wm title $top "Alicq chat"
+		Toplevel $top AlicqUserWindow
 		grid [frame $top.btn -class Tablist] -sticky new -row 0
 		grid columnconfigure $top 0 -weight 1
 		grid rowconfigure $top 1 -weight 1
@@ -140,6 +155,7 @@ proc SelectTab {top name args} {
 		}
 	}
 	grid $win -sticky news -row 1
+	ChatTitle $top $uid "Alicq chat"
 	focus $win
 }
 
@@ -164,6 +180,7 @@ handler *|TaggedText highlightTab {uid args} {
 
 # Common dialog frame used ni both single and tabbed modes
 proc DialogFrame {top uid} {
+	variable windata
 	foreach {chat in out} [ChatBar $top.pw] break
 	grid $chat -sticky news
 	grid [EditBar $top.edit $uid $in $out] -sticky we
@@ -173,9 +190,28 @@ proc DialogFrame {top uid} {
 		lappend ids [hook $x|Acknowledgement [nc AckMessage $in]]\
 			    [hook $x|TaggedText [nc DisplayMessage $in] 0.9]
 	}
-	bind $top <FocusIn> [list focus $out]
-	bind $in <Destroy> [nc unhook $ids]
+	bind $top <FocusIn> [nc DoFocus %W $out]
+	set format [option get $in messageStyle MessageStyle]
+	if {[set fmt [lindex [info commands header::$format] 0]]==""} {
+		set fmt header::default
+	}
+	array set windata [list $in,uid $uid $in,hooks $ids $in,format $fmt]
+	bind $in <Destroy> [nc DialogClose $in $out]
 	foreach x [list $top $out] { bind $x <<Close>> [list destroy $top] }
+	if {[set fmt [lindex [info commands display::$format] 0]]==""} {
+		set fmt display::default
+	}
+	$fmt $in
+	Event ChatOpen $uid $in $out
+}
+
+proc DoFocus {from to} { if {[winfo toplevel $from]==$from} { focus $to } }
+
+proc DialogClose {in out} {
+	variable windata
+	unhook $windata($in,hooks)
+	Event ChatClose $windata($in,uid) $in $out
+	array unset windata $in,*
 }
 
 # Editbar contains buttons to control message edition and sending
@@ -188,13 +224,13 @@ proc EditBar {name uid in out} {
 	# Check if uid supports several message types
 	set types [Event MessageTypes $uid]
 	if {[llength $types]} {
-		TypeButton $name.type $uid $types $name.send
-		grid $name.type	-row 0 -column 1 -sticky w
+		TypeButton $name.type $uid $types $name.send $out
+		grid $name.type	-row 0 -column 2 -sticky w
 	}
 	if {[winfo class [winfo parent $name]]!="Tab"} { 
 		CompoundSend $name.send $uid 
 	}
-	grid [button $name.close] -row 0 -column 10 -sticky e
+	grid [button $name.close] -row 0 -column 10 -sticky e -padx 4
 	$name.close configure -command [list event generate [winfo parent $name] <<Close>>]
 	frame $name.aux -class AuxBar
 	# If multicast mode, add contact aliases to citations
@@ -205,14 +241,16 @@ proc EditBar {name uid in out} {
 	if {!$multi} {
 		button $name.aux.history
 		HistoryButton $name.aux.history $uid [Event HistoryChunks $uid]
+		bind $out <<History>> [list $name.aux.history invoke]
 		grid $name.aux.history -row 0 -column 2 -sticky wns
+		checkbutton $name.dc -indicatoron 0 -text ""
+		#grid $name.dc -row 0 -column 1 -sticky ns
 	}
-	if {[llength [info command tooltip]]} { 
-		tooltip $name.aux.cite "Cite last message" <<Cite>>
-		tooltip $name.close "Close dialog" <<Close>>
-		tooltip $name.send "Send message" <<Send>>
-	}
-	grid $name.aux -sticky wns -row 0 -column 2 -ipadx 1 -padx 5
+	ui::tooltip $name.aux.cite "Cite last message" <<Cite>>
+	ui::tooltip $name.close "Close dialog" <<Close>>
+	ui::tooltip $name.send "Send message" <<Send>>
+	if {!$multi} {ui::tooltip $name.aux.history "Show history" <<History>>}
+	grid $name.aux -sticky wns -row 0 -column 3 -ipadx 1 -padx 5
 	foreach x {1 2 3} { grid columnconfigure $name $x -weight 1}
 	set name
 }
@@ -235,24 +273,22 @@ handler MessageTypes grouptypes {uid} {
 
 # Create menubutton for selection of message type from a list
 # Selected type is stored in variable with UID-based name
-proc TypeButton {name uid types send} {
+proc TypeButton {name uid types send out} {
 	set var [namespace current]::${uid}-type
-	menubutton $name -menu $name.vals
-	menu $name.vals -tearoff no
-	set max 0
-	foreach x $types {
-		set t [mc $x]
-		$name.vals add radiobutton -label $t -value $x -variable $var\
-			-command [list $name configure -text $t]
-		set len	[string length $t]
-		if {$len>$max} {set max $len}	
-	}
-	$name configure -width $max
-	bind $name.vals <Destroy> [list unset $var] 
-	$name.vals invoke 0
+	set values [map x $types {list $x [mc $x]}]
+	ui::variant $name -variable $var -values $values 
+	trace variable $var w [nc TypeSelected $uid $out]
+	bind $name.menu <Destroy> [list + unset $var] 
+	$name.menu invoke 0
 	$send configure -command [join [list [$send cget -command]\
-		[list $name.vals invoke 0]] "\n"]
-	set name	
+		[list $name.menu invoke 0]] "\n"]
+	set name
+}
+
+proc TypeSelected {uid txt name args} {
+	upvar 1 $name type
+	set cmd [namespace current]::Type:$type
+	if {[llength [info procs $cmd]]} { $cmd $uid $txt }
 }
 
 # Chatbar is panned window with text widgets for incoming and outgoing messages
@@ -261,9 +297,9 @@ if {[package vsatisfies [package present Tk] 8.4]} {
 	option add *Panedwindow*in.txt.state disabled
 	option add *Framein.height 200 widgetDefault
 	option add *Frameout.height 100 widgetDefault
-	option add *EditBar.AuxBar.cite.image cite widgetDefault
+	option add *EditBar.AuxBar.cite.image button:cite widgetDefault
 
-	image create bitmap cite -foreground brown -data {
+	image create bitmap button:cite -foreground brown -data {
 		#define cite_width 16
 		#define cite_height 16
 		static unsigned char cite_bits[] = {
@@ -310,6 +346,7 @@ if {[package vsatisfies [package present Tk] 8.4]} {
 } else {
 	option add *PanedWindow.f0*txt.state disabled
 	option add *Frame.height 200 widgetDefault
+	option add *EditBar.close.text x widgetDefault
 	proc ChatBar {n} {
 		PanedWindow $n -side right -activator line
 		foreach x {in out} {set $x [MFrame [$n add -weight 1] $x]}
@@ -327,7 +364,7 @@ if {[package vsatisfies [package present Tk] 8.4]} {
 }
 
 proc MFrame {fr x} {
-	bindtags $fr [concat MsgFrame Frame$x [bindtags $fr]]
+	bindtags $fr [concat MsgFrame Frame$x [bindtags $fr] SaveProportions]
 	[ScrolledWindow $fr.sw] setwidget [text $fr.txt -wrap word -bd 1]
 	bindtags $fr.txt [concat MsgText Text$x [bindtags $fr.txt]]
 	grid $fr.sw -sticky news
@@ -372,6 +409,7 @@ proc SendMessage {uid txt} {
 	set m $msg
 	if {[llength [info commands Form$type]]} {set m [Form$type $msg]}
 	set tags [concat outgoing $type [Event Send $type $uid $m]]
+	#Event $uid|incoming:$type $uid [clock seconds] [list $msg] $tags
 	Event $uid|TaggedText Me [clock seconds] [list $msg] $tags
 	$txt delete 1.0 end
 }
@@ -391,69 +429,167 @@ proc FormURL {msg} {
 }
 
 proc DisplayMessage {txt uid time message {tags {}} {pos end} {see 1}} {
+	variable windata
 	$txt configure -state normal
 	$txt tag delete last
-	set fmt [option get $txt timeFormat TimeFormat]
-	set stime "\[[clock format $time -format $fmt]\] "
-	set alias [get $uid Alias]
-	lappend tags $uid
-	set message [concat [list $stime {header time} $alias {header alias}\
-		":\n" header] $message]
 	set msg [list]
-	foreach {part t} $message {lappend msg $part [concat $t $tags last]}
-	eval [list $txt insert $pos] $msg {\n\n}
-	event generate $txt <<Message>>
-	if {$see} {$txt see end}
+	foreach {part t} $message {lappend msg $part [concat $t body]}
+	set header [list]
+	foreach {part t} [$windata($txt,format) $uid $time $txt] {
+		lappend header $part [concat $t header]
+	}
+	set message [concat $header $msg]
+	set msg [list]
+	lappend tags $uid time:$time
+	foreach {part t} $message {lappend msg $part [concat $t $tags]}
+	set msg [concat [lrange $msg 0 end-2] [list [string trimright\
+			[lindex $msg end-1] "\n "] [lindex $msg end] "\n" ""\
+			"\n" delimiter]]
+	set from [$txt index $pos]		
+	set fraction [lindex [$txt yview] end]
+	eval [list $txt insert $from] $msg 
+	foreach {from end} [$txt tag nextrange body $from] break
+	if {[info exists end] && $end!=""} {
+		Event DisplayMessage $txt $from $end
+	}
+	if {$fraction > 0.99 || $see} {$txt see end}
 	$txt configure -state disabled
 	delayed [nc Raise [winfo toplevel $txt]]
 	set txt
 }
 
-proc NewObjectDialog {} {
-	set top .new-obj
+proc Time {txt time} {
+	set fmt [option get $txt timeFormat TimeFormat]
+	if {[clock seconds]-$time>12*3600} { 
+		set fmt "[option get $txt dateFormat DateFormat] $fmt"
+	}
+	clock format $time -format $fmt
+}
+
+namespace eval header {
+	proc default {uid time txt} {
+		upvar #0 [namespace parent]::windata win
+		set stime "\[[[namespace parent]::Time $txt $time]\] "
+		set alias [get $uid Alias]
+		list $stime time $alias alias ":\n" ""
+	}
+	proc Chat {uid time txt} {
+		set stime " ([[namespace parent]::Time $txt $time])"
+		set alias [get $uid Alias]
+		list $alias alias $stime time ": " ""
+	}
+}
+
+namespace eval display {
+	proc default {txt} {
+	}
+}
+
+handler AddItem AddItem {class args} {
+	set top .add-[string map {: -} $class]
 	if {[winfo exists $top]} {
-		raise $top
-		return
+		raise $top; focus $top; return
 	}
-	toplevel $top -class NewObject
-	grid [frame $top.content] -sticky news -padx 2 -pady 2
-	grid [frame $top.btn] -sticky we -padx 2
-	foreach x {row column} { grid ${x}configure $top 0 -weight 1 }
-	grid columnconfigure $top.content 1 -weight 1
-	button $top.btn.ok -text [mc Create] -default active
-	button $top.btn.cancel -text [mc Cancel] -command [list destroy $top]
-	grid $top.btn.ok $top.btn.cancel -sticky w -padx 10 -pady 2
-	New:Contact $top.content $top.btn.ok
+	foreach {domain type} [split $class :] break
+	Toplevel $top AlicqAddItem
+	wm title $top "Add new $type [string tolower $domain]"
+	set meta [ref $class]::meta
+	set cmd ""
+	if {[info exists ${meta}::Identifier]} {
+		set name [set ${meta}::Identifier]
+		if {[info commands ::validate::$name]!=""} { 
+			set cmd ::validate::$name
+		}
+	} else { set name Identifier }
+	set var [namespace current]::$top
+	grid [label $top.prompt] -sticky we -padx 8 -padx 2 -columnspan 2
+	if {[info exists ${meta}::AddPrompt]} {
+		$top.prompt configure -text [set ${meta}::AddPrompt]
+	}
+	grid [label $top.lbId -text [mc $name]]\
+	     [entry $top.enId -validate key -vcmd\
+	     		[nc ValidId $class $cmd $top.btn.add %P]]\
+		-padx 8 -pady 4 -sticky new
+	grid [label $top.lbAlias -text [mc Alias]] [entry $top.enAlias]\
+		-padx 8 -pady 4 -sticky new
+	grid [label $top.lbGroups -text [mc Groups]]\
+		-columnspan 2 -padx 8 -pady 4 -sticky we
+	grid [ScrolledWindow $top.sw]\
+		-row 8 -columnspan 2 -padx 8 -pady 4 -sticky news
+	$top.sw setwidget [Tree $top.trGroups -showlines 0 -deltay 20]
+	grid [frame $top.btn -class ButtonBar] -row 11 -columnspan 2\
+		-padx 8 -pady 4 -sticky wes
+	button $top.btn.add -text [mc Add] -state disabled\
+		-command [nc AddItemRun $class $top.enId $top.enAlias $var]
+	button $top.btn.close -command [list destroy $top]
+	grid $top.btn.add $top.btn.close -padx 4 -pady 2 -sticky e
+	grid columnconfigure $top 1 -weight 1
+	grid rowconfigure $top 8 -weight 1
+	
+	foreach {id alias groups} $args break
+	if {[info exists id] && $id!=""} { $top.enId insert 0 $id }
+	if {[info exists alias] && $alias!=""} { $top.enAlias insert 0 $alias }
+	if {[info exists groups]} { foreach x $groups { set ${var}($x) 1} }
+
+	GroupTree $top.trGroups $var
+	set top
 }
-proc New:Contact {top ok} {
-	wm title [winfo toplevel $top] "New ICQ contact"
-	foreach x {"ICQ UIN" "Alias"} {
-		label $top.lb$x -text [mc $x] -anchor e
-		entry $top.en$x
-		grid $top.lb$x $top.en$x -sticky we -padx 4 -pady 4
+bind AlicqAddItem <Destroy> [nc unset %W]
+
+# Prepare groups for recursive tree population
+proc GroupTree {tree var} {
+	array set hier [list]
+	array set $var [list]
+	foreach x [select Group:common] {
+		set id [lindex [split $x :] end]
+		if {$id=="other"} continue
+		set groups [get $x Groups]
+		if {$groups==""} { set groups ":" }
+		foreach g $groups { lappend hier($g) $id }
+	}
+	variable nodecounter
+	set nodecounter 0
+	PopulateTree $tree hier $var root :
+	unset hier nodecounter
+}
+
+# Recursively populate group tree, taking prepared information from data array
+proc PopulateTree {tree data var parent node} {
+	upvar 1 $data hier
+	variable nodecounter
+	if {![info exists hier($node)]} return
+	foreach x $hier($node) {
+		set name [set [ref Group:common:$x](Alias)]
+		set node [$tree insert end $parent [incr nodecounter]\
+			-data $x -open 1]
+		$tree itemconfigure $node -window [checkbutton $tree.c$node\
+			-text $name -variable ${var}($x)]
+		PopulateTree $tree hier $var $node $x
 	}
 }
 
-# New objects dialogs
-hook add:group [namespace code {
-	NewObjDialog group ::AddObject {Enter group name:}
-	}]
-hook add:contact [namespace code {
-	NewObjDialog contact ::AddObject {Enter UIN:}
-	}]
+# Validate new object ID against its type, if given. Also check if object
+# with same identifier already exists
+proc ValidId {class cmd btn val} {
+	if {$val=="" || [info exists [ref ${class}:${val}]]} {
+		$btn configure -state disabled
+		return 1
+	} elseif {[set res [expr {($cmd=="")?1:[$cmd $val]}]]} {
+		$btn configure -state normal
+	}
+	set res
+}
 
-proc NewObjDialog {type action caption args} {
-	#NewObjectDialog
-	set top ".new-$type"
-	if {[catch {toplevel $top}]} return
-	wm title $top "New $type"
-	label $top.lb -text $caption
-	entry $top.en
-	pack $top.lb $top.en -fill x
-	append action " $type \[$top.en get\]; destroy $top"
-	pack [ButtonBar $top.buttons [list\
-		[list Create $action <<Accept>>]\
-		[list Close "destroy $top" <<Close>>]]] -side bottom
+# Get information for AddItem dialog and create new object
+proc AddItemRun {class enId enAlias var} {
+	set id [$enId get]
+	set data(Alias) [$enAlias get]
+	if {$data(Alias)==""} { unset data(Alias) }
+	foreach {key val} [array get $var] { 
+		if $val {lappend data(Groups) $key} 
+	}
+	destroy [winfo toplevel $enId]
+	new $class:$id [array get data]
 }
 
 proc AddIncomingContacts {list} {
@@ -487,16 +623,20 @@ handler {Contact:ICQ:*|Send:contacts} SendContacts {id} {
 	set uid [lindex [split $id |] 0]
 	set top [WinByUid contacts $uid]
 	if {![catch {raise $top}]} return
-	toplevel $top -class AlicqContactsWindow
+	Toplevel $top AlicqContactsWindow
 	wm title $top "Contacts for $uid"
 	set lbox [ListBox $top.list -deltay 20 -dropenabled yes\
 		-droptypes {TREE_NODE} -dropcmd [nc Drop]]
-	$lbox insert end zz -fill blue -text "Drad contacts to send here"
+	$lbox insert end zz -fill blue -text [mc "Drag contacts to send here"]
 	grid $lbox -sticky news
-	grid [ButtonBar $top.btn [list\
-		[list Send "Event Send contacts $uid \[[nc SendContactlist $lbox]\];destroy $top" <<Send>>]\
-		[list Cancel "destroy $top"]\
-	]]
+	grid [frame $top.btn2 -class EditBar] -sticky wes -padx 4
+	grid [button $top.btn2.send -command [nc SendContactlist $uid $lbox]]\
+			-sticky nws -row 1 -column 1
+	grid [button $top.btn2.close -command [list destroy $top]]\
+		-sticky nes -row 1 -column 5
+	bind $top <<Send>> [list $top.btn2.send invoke]	
+	bind $top <<Close>> [list $top.btn2.close invoke]	
+	grid columnconfigure $top.btn2 4 -weight 1
 	grid rowconfigure $top 0 -weight 1
 	grid columnconfigure $top 0 -weight 1
 }
@@ -505,12 +645,13 @@ handler {*|Send:authorization} authorization {id} {
 	Event Send authorization [lindex [split $id |] 0] grant
 }
 
-proc SendContactlist {name} {
+proc SendContactlist {recipient name} {
 	set res [list]
 	foreach uid [$name items 1 end] {
 		lappend res [lindex [split $uid :] 2] [get $uid Alias]
 	}
-	set res
+	if {[llength $res]} { Event Send contacts $recipient $res }
+	destroy [winfo toplevel $name]
 }
 
 proc Drop {widget src dest op datatype data} {
@@ -602,9 +743,7 @@ proc AliasChanged {uid ref field args} {
 	if {[string is true $tabbing]} {
 		set win .all.btn$win
 		if {[winfo exists $win]} { $win configure -text $alias }
-	} elseif {[winfo exists $win]} {
-		wm title $win "$alias \([lrange [split $uid :] 1 end]\)"
-	}
+	} elseif {[winfo exists $win]} { ChatTitle $win $uid "Alicq" }
 }
 
 # Transform Incoming message to typed incoming:type messages
@@ -622,9 +761,8 @@ proc retransform {flag type uid time message msgid} {
 	Event $uid|incoming:$type $uid $time $message $msgid
 }
 
-
-handler *|incoming:text tagtext {uid time message id} { 
-	Event $uid|TaggedText $uid $time [list $message {}] [list incoming $id]
+handler *|incoming:text tagtext {uid time message id {t incoming}} { 
+	Event $uid|TaggedText $uid $time [list $message {}] [concat $t $id]
 } 0.90
 
 handler *|incoming:URL tagurl {uid time msg msgid} {
@@ -652,18 +790,21 @@ handler *|show:contacts ContactsDialog {uid args} {
 	set top ".contacts-$UIN"
 	if {![catch {raise $top}]} return
 
-	toplevel $top -class AlicqContactsWindow
+	Toplevel $top AlicqContactsWindow
 	wm title $top "Contacts from [get $uid Alias] ($UIN)"
 
 	grid [ListBox $top.list -dragenabled 1] -sticky news
 	$top.list bindText <1> [list $top.list selection add]
 	$top.list bindText <3> [list $top.list selection remove]
-	grid [ButtonBar $top.btn [list\
-		[list Add [nc AddIncomingContacts $top.list]]\
-		[list Close "destroy $top" <<Close>>]\
-	]]
+	grid [frame $top.btn -class ButtonBar] -sticky news -pady 2
+	button $top.btn.add -command [nc AddIncomingContacts $top.list]\
+		-text [mc Add]
+	button $top.btn.close -command [list destroy $top]
+	grid $top.btn.add x $top.btn.close -sticky e -padx 4
+	bind $top <<Close>> [list $top.btn.close invoke]
 	grid rowconfigure $top 0 -weight 1
 	grid columnconfigure $top 0 -weight 1
+	grid columnconfigure $top.btn 1 -weight 1
 	set id [hook $uid|incoming:contacts [nc DisplayContacts $top.list] 0.9]
 	bind $top.list <Destroy> [list unhook $id]
 }
@@ -695,13 +836,45 @@ proc AuthorizeFromDialog {txt} {
 	if {[set pos [lsearch $tags action:*]]!=-1} {
 		set action [lindex [split [lindex $tags $pos] :] end]
 	} else return
-	Event Send authorization $uid $action
+	set tags [Event Send authorization $uid $action]
+	Event $uid|TaggedText Me [clock seconds] [list $action] $tags
 }
 
-proc CloseButton {name} {
-	button $name
-	$name configure -command [list destroy [winfo parent $name]]
-	if {[$name cget -label]==""} { $name configure -label x}
+namespace eval display {}
+
+proc display::basic {win} {
+	# Setup colors
+	foreach x {incoming outgoing sent client server action authrequest} {
+		set val [option get $win ${x}Color [string totitle $x]Color]
+		if {$val!=""} { $win tag configure $x -foreground $val }
+	}
+	foreach {x y} {alias ContactProperties action AuthorizeFromDialog} {
+		$win tag bind $x <Enter> { %W configure -cursor hand2 }
+		$win tag bind $x <Leave> { %W configure -cursor {} }
+		$win tag bind $x <1> [list [namespace parent]::$y %W]
+	}
+	$win tag configure action -underline 1
+}
+
+proc display::default {win} {
+	basic $win
+	# Underline given alias if needed
+	set underline [option get $win underlineTag UnderlineTag]
+	if {$underline!=""} { $win tag configure $underline -underline yes}
+
+	# Configure header options
+	foreach x {background font relief} {
+		set opt header[string totitle $x]
+		set val [option get $win $opt [string totitle $opt]]
+		if {$val!=""} { $win tag configure header -$x $val}
+	}
+}
+
+proc display::Chat {win} {
+	basic $win
+	$win tag configure body -foreground [option get $win bodyColor BodyColor]
+	$win tag raise body
+	$win tag configure delimiter -elide yes
 }
 
 # Event bindings
@@ -710,30 +883,30 @@ bind MsgFrame <Destroy> {+
 }
 
 bind Textin <Map> [namespace code {
-	# Setup colors
-	foreach x {incoming outgoing sent client server action authrequest} {
-		set val [option get %W ${x}Color [string totitle $x]Color]
-		if {$val!=""} { %W tag configure $x -foreground $val }
-	}
-	# Underline given alias if needed
-	set underline [option get %W underlineTag UnderlineTag]
-	if {$underline!=""} { %W tag configure $underline -underline yes}
-
-	# Configure header options
-	foreach x {background relief font} {
-		set opt header[string totitle $x]
-		set val [option get %W $opt [string totitle $opt]]
-		if {$val!=""} { %W tag configure header -$x $val}
-	}
-	foreach {x y} {alias ContactProperties action AuthorizeFromDialog} {
-		%W tag bind $x <Enter> { %W configure -cursor hand2 }
-		%W tag bind $x <Leave> { %W configure -cursor {} }
-		%W tag bind $x <1> [namespace code [list $y %W]]
-	}
-	%W tag configure action -underline 1
 	%W tag bind incoming <3> {.incoming post %%X %%Y}
 }]
 
+if {$::tcl_platform(platform)=="windows"} {
+	bind Textin <ButtonPress-1> { 
+		if {![info exists [list focus-%W]] && [focus]!=[list %W]} {
+			set [list focus-%W] [focus]
+			focus %W
+		}
+	}
+
+	bind Textin <KeyPress> { 
+		if {[string is alnum "%A"]} {
+			focus [set [list focus-%W]]
+			event generate [set [list focus-%W]] <KeyPress>\
+				-keycode %k -keysym %K
+		}
+	}
+
+	bind Textin <FocusOut> {
+		if {[info exists [list focus-%W]]} { unset [list focus-%W] }
+	}
+}
+
 bind Textout <Map> {
 	%W tag configure citation -foreground brown
 }
