@@ -2,6 +2,7 @@
 
 package require icq 0.8.8
 
+# Module metainformation '''1
 namespace eval meta {
 	set name "ICQ"
 	set author "Ihar Viarheichyk <iverg@mail.ru>"
@@ -25,9 +26,13 @@ namespace eval meta {
 		default [encoding system] desciption "Default encoding"\
 		property Global:ICQ:Info|encoding weight .13]
 	unset x	
-	array set ChangePassword {type action menu {ICQ "Change Password"}}
+	array set ChangePassword {type action menu {Change "ICQ Password"}}
+	array set add:contact {type action menu {Add Contact}
+		script { Event AddItem Contact:ICQ }
+	}
 }
 
+# Network metainformation '''2
 namespace eval network::meta {
 	array set server {
 		type fqdn default login.icq.com save change
@@ -39,9 +44,8 @@ namespace eval network::meta {
 	}
 	array set proxy {
 		type variant save change default "" weight .23
-		description "Proxy to use" menu "ICQ Proxy"
-		valuescript {concat {""} [map x [select Proxy] {
-				lindex [split $x :] end}]}
+		description "Proxy to use" empty "No proxy"
+		valuescript {map x [select Proxy] { lindex [split $x :] end}}
 	}
 	array set reconnect {
 		type boolean save change property Global:ICQ:Network|reconnect
@@ -58,6 +62,7 @@ namespace eval network::meta {
 	}
 }
 
+# Roster metainformation '''2
 namespace eval roster::meta {
 	array set use { 
 		type boolean default 1 property Global:ICQ:Roster|use
@@ -78,9 +83,25 @@ namespace eval roster::meta {
 		save change default 0
 	}
 }
+#```1
 
-# ICQ-specific properties for ICQ contacts
+if {[catch { package require direct 0.3 } reason]} {
+	Event Log warning "Direct connections are disabled: $reason"
+	namespace eval direct { set enable 0 }
+} else {
+	namespace eval direct::meta {
+		array set enable {
+			type boolean default 0 save change
+			description "Enable direct connections"
+		}
+	}
+}
+
+# ICQ-specific properties for ICQ contacts '''1
 namespace eval [ref Contact:ICQ]::meta {
+	set Identifier UIN
+	set AddPrompt { You should specify numerical UIN of ICQ contact. 
+	}
 	array set Status  {
 		type {cache variant} default offline 
 		values {offline online occ dnd ffc away na invisible}
@@ -107,16 +128,18 @@ array set ${ref}::meta::encoding [array get meta::encoding]
 # Sync contact default encoding with global encoding
 upvar #0  ${ref}::meta::encoding(default) [namespace current]::encoding
 
-
 namespace eval [ref Group:common]::meta {
 	array set rid	{type integer save change}
 }
 
+# Data validation functions '''1
 namespace eval ::validate {
 	proc port {val} {
 		expr { [string is integer $val] && $val>=0 && $val<=65535 }
 	}
+	proc UIN {val} { string is digit $val }
 }
+# ```1
 
 variable icq
 variable recreate 0
@@ -124,17 +147,51 @@ variable itemCount 0
 
 proc Uid2Uin {list} { map x $list { lindex [split $x :] end } }
 
-# Proceed ICQ events
+# Process ICQ events '''1
 proc ICQEvent {event args} {
 	if {[llength [info commands icq:$event]]} {
 		eval icq:$event $args
 	} else { eval Event [list $event] $args}
 }
 
+proc icq:Log {tags str} { Event Log [concat $tags ICQ] $str }
 # Process contact statuses
-proc icq:Status {uin newstatus {ip ""}} {
+proc icq:Status {uin newstatus} {
 	set [ref Contact:ICQ:$uin](Status) $newstatus
-	if {$ip!=""} { set [ref Contact:ICQ:$uin](IP) $ip }
+}
+
+proc icq:IP {uin ip} {
+	if {$ip!="0.0.0.0"} { set [ref Contact:ICQ:$uin](IP) $ip }
+}
+
+proc icq:LanDetails {uin type ip port cookie version} {
+	set ref [ref Contact:ICQ:$uin]
+	if {$ip!=[set ${ref}(IP)] && $ip!="0.0.0.0"} {set ${ref}(LocalIP) $ip}
+	set ${ref}(DC:type) $type
+	# If DC is disabled, delete previous instance, if present
+	if {$type=="disabled" || $type=="web" || $ip=="0.0.0.0" || $port==0} {
+		if {[info exists ${ref}(dc)]} { 
+			[set ${ref}(dc)] delete
+			unset ${ref}(dc) 
+		}
+		return
+	}
+	# If no previous instance found, create new DC handler
+	if {![info exists ${ref}(dc)]} { 
+		set ${ref}(dc) [::direct::peer $uin -event [nc DcEvent]]
+	}
+	variable encoding
+	if {[info exists ${ref}(encoding)]} {
+		set enc [set ${ref}(encoding)]
+	} else { set enc $encoding }
+	variable dcserver
+	if {[info exists dcserver]} {
+		set mp [lindex [$dcserver info] 0]
+	} else { set mp 0 }
+	[set ${ref}(dc)] configure -host $ip -port $port -cookie $cookie\
+		-uin [set [namespace current]::uin] -encoding $enc -myport $mp
+	Event Log info "DC connection to $uin [[set ${ref}(dc)] state]"
+	Event Log warning "DC cookie for $uin $cookie"
 }
 
 # Process personal status
@@ -142,24 +199,38 @@ proc icq:MyStatus {newstatus} {
 	if {$newstatus=="offline"} {
 		foreach x [select Contact:ICQ] { set [ref $x](Status) offline }
 	} 
-	set [ref Me](Status) $newstatus
+	upvar 0 [ref Me](Status) status
+	if {$status=="offline" && $newstatus!="offline"} onOnline
+	set status $newstatus
 }
 
 # Translate ICQ message to alicq messages
 proc icq:Incoming {type sender time message} {
 	set id [clock seconds]:[clock clicks] 
-	if {$type!="web"} {
-		Event Incoming $type Contact:ICQ:$sender $time $message $id
-	} else {
+	set message [string trimright $message "\0"]
+	if {$type=="web"} {
 		Event Incoming text Contact:ICQ:0 $time "$sender:\n$message" $id
+	} else {
+		if {$type=="auto"} { 
+			set type "text"
+			set [ref Contact:ICQ:$sender](description) $message
+		} elseif {$type=="authorization" && $message=="granted"} {
+			Event Log notice "autorization granted by $sender"
+			set [ref Contact:ICQ:$sender](auth) 0
+		}
+		Event Incoming $type Contact:ICQ:$sender $time $message $id
 	}
 }
 
+proc icq:MyIP {local remote} {
+	Event Log info "Local IP $local, remote IP $remote"
+}
+
 proc icq:Registered {uin password} {
 	variable recreate 1
 	set [namespace current]::uin $uin
 	set [namespace current]::password $password
-	Event SetStatus online
+	after idle [nc Event SetStatus online]
 }
 
 proc icq:RegistrationRefused {} {
@@ -171,16 +242,22 @@ proc icq:Outgoing {type sender time message msgid} {
 }
 
 proc icq:Client {uin args} {
-	set [ref Contact:ICQ:$uin](Client) $args
+	set ref [ref Contact:ICQ:$uin]
+	set ${ref}(Client) $args
+	# If client supoprts unicode, use UTF-8 for direct connections
+	if {[info exists ${ref}(dc)] && [lindex $args end]} {
+		[set ${ref}(dc)] configure -encoding utf-8
+	}
 }
 
 proc icq:ACK {type uin msgid} {
-	Event Log 1 "Ack $type from $uin on $msgid"
+	Event Log info "Ack $type from $uin on $msgid"
 	Event Contact:ICQ:$uin|Acknowledgement $type Contact:ICQ:$uin $msgid
 }
 
 proc icq:Capabilities {uin caps} {
-	Event Log 5 "Capabilities $uin: $caps"
+	set [ref Contact:ICQ:$uin](Capabilities) $caps
+	Event Log info "Capabilities $uin: $caps"
 }
 
 proc icq:Roster:Items {items time} {
@@ -192,7 +269,7 @@ proc icq:Roster:Items {items time} {
 			continue
 		}
 		if {[llength [info commands roster:$type]]} {
-			Event Log 8 "roster $type: $id $name $aux"
+			Event Log debug "roster $type: $id $name $aux"
 			if {[catch { roster:$type $id $name $aux } v]} {
 				Event Error roster "In roster $type $name: $v"
 			}
@@ -209,10 +286,78 @@ proc icq:Roster:Items {items time} {
 proc icq:Roster:OK {args} { MonitorRoster }
 
 proc icq:Roster:Update {status} {
-	Event Log 6 "Contact updated with status $status"
+	Event Log debug "Contact updated with status $status"
 	ProcessResult $status
 }
 
+proc icq:Missed {uin reason channel} {
+	set uid Contact:ICQ:$uin
+	Event Log {error missed} "Missed message from $uin: $reason"
+	set auto [mc "Message from this contact was discarded by server."]
+	if {[string is integer $reason]} {
+		append auto "\n" [mc "The description of the reason is not available, please contact autor of Alicq and report error code."] \
+			"\n" [mc "Error code"] " " $reason
+	} else { append auto "\n" [mc "Reason:"] " " [mc $reason] }
+	Event Incoming text $uid [clock seconds] $auto 0
+}
+
+# Direct connection events '''1
+proc DcEvent {uin event args} {
+	set uid Contact:ICQ:$uin
+	if {[llength [info commands DC:$event]]} {
+		eval [list DC:$event $uid] $args
+	} else { Event Log warning "Unhandler DC event $event from $uid" }
+}
+
+proc DC:Log {uid tags str} { Event Log [concat $tags ICQ:DC] "Peer $uid: $str"}
+
+proc DC:Disconnected {uid reason} {
+	set ref [ref $uid]
+	Event Log {info DC} "DC for $uid is closed: $reason"
+	if {[info exists ${ref}(_is_open)]&&[set ${ref}(Status)]!="offline"} {
+		Event Log {info DC} "reopening again"
+		after 1000 [list [set ${ref}(dc)] connect]
+	}
+}
+
+proc DC:Outgoing {uid type message time msgid} {
+	Event Outgoing $type $uid $time $message $msgid
+}
+
+proc DC:Incoming {uid type message} {
+	Event Log info "Peer incoming $type from $uid"
+	set id [clock seconds]:[clock clicks] 
+	Event Incoming $type $uid [clock seconds]\
+		[string trimright $message "\0"] $id 
+}
+
+proc DC:ACK {uid type id} { Event $uid|Acknowledgement $type $uid $id }
+
+proc DC:Timeout {uid type message id} {
+	variable icq
+	Event Log notice "DC timoeout, resending via server"
+	$icq send $type [lindex [split $uid :] 2] $message $id
+}
+#}}}1
+# Turn DC server on/off
+proc MonitorDCServer {args} {
+	variable dcserver
+	variable uin
+	variable icq
+	set dcinfo [list]
+	if {$direct::enable && ![info exists dcserver]} {
+		if {[catch { 
+			set dcserver [::direct::server 0 -uin $uin\
+					-event [nc DcEvent]]
+			set dcinfo [$dcserver info]  } r]
+		} { Event Log error $r } else { Event Log info "DC server $dcserver"}
+	} elseif {!$direct::enable && [info exists dcserver]} {
+		$dcserver delete
+		unset dcserver
+	}
+	if {[info exists icq]} { $icq configure -dcinfo $dcinfo }
+}
+
 # Install server-side roster related monitoring of objects
 proc MonitorRoster {args} {
 	variable groups
@@ -260,7 +405,7 @@ proc Move {uid args} { foreach x {delete add} {Enqueue $x $uid} }
 proc Enqueue {action uid args} { 
 	if {![string is true $roster::modify]} return
 	lappend roster::queue [list $action $uid] 
-	Event Log 6 "enqueue roster item $action for $uid"
+	Event Log debug "enqueue roster item $action for $uid"
 	ProcessQueue
 }
 
@@ -278,12 +423,15 @@ proc ProcessQueue {} {
 # Handle result of SSI operation, and shedule processing of new item in
 # roster actions queue if needed
 proc ProcessResult {result} {
+	variable icq
 	variable rosterLock
 	if {![info exists rosterLock] || ![llength $roster::queue]} return
 	unset rosterLock
 	foreach {action uid} [lindex $roster::queue 0] break
 	# Contact is not authorized, try to add it with auth TLV
 	if {$result==14} {
+		Event Send authrequest $uid {I want to add you in contact list}
+		Event Send futureauth $uid {}
 		set [ref $uid](auth) 1
 	} else {
 		if {$result==0} {
@@ -310,7 +458,7 @@ proc RosterSync {action uid} {
 	set rid [Roster:$action $uid]
 	foreach {type name aux} [RosterInfo $uid] break
 	set list [list $action [list $type $rid $name $aux]]
-	Event Log 6 [list $icq roster $action [list $type $rid $name $aux]]
+	Event Log debug [list $icq roster $action [list $type $rid $name $aux]]
 	$icq roster $action [list $type $rid $name $aux]
 	return 1
 }
@@ -419,7 +567,6 @@ proc EstimatedRoster {} {
 	       ${roster::count} }
 }
 
-
 proc CreateConnection {{new 0} {local_password ""}} {
 	foreach x {recreate alias uin password} { variable $x }
 	set recreate 0
@@ -437,11 +584,14 @@ proc CreateConnection {{new 0} {local_password ""}} {
 		-capabilities {ICQ RELAY UTF8}\
 		-event [namespace current]::ICQEvent]
 	foreach x [info vars network::*] {
-		$icq configure -[namespace tail $x] [set $x]
+		set val [set $x]
+		if {[catch { $icq configure -[namespace tail $x] $val } r]} {
+			Event Log error "Cant configure [namespace tail $x]: $r"
+		}
 	}
 	if {$local_uin!="new" && [string is true $roster::use]} {
 		if {$roster::updated>0} {
-			Event Log 6 "Estimated roster: [EstimatedRoster], $roster::updated"
+			Event Log info "Estimated roster: [EstimatedRoster], $roster::updated"
 			set roster [list $roster::updated [EstimatedRoster]]
 		} else {set roster 1}
 	} else {set roster 0}
@@ -458,6 +608,8 @@ proc CreateConnection {{new 0} {local_password ""}} {
 	foreach x [select Contact:ICQ {[info exists encoding]}] {
 		$icq encoding [set [ref $x](encoding)] [Uid2Uin $x]
 	}
+	variable dcserver
+	if {[info exists dcserver]} { $icq configure -dcinfo [$dcserver info] }
 	set icq
 }
 
@@ -517,6 +669,9 @@ proc trackencoding {uin ref field args} {
 	variable icq
 	upvar 1 ${ref}($field) val
 	$icq encoding $val $uin
+	if {[info exists ${ref}(dc)]} { 
+		[set ${ref}(dc)] configure -encoding $val
+	}
 }
 
 # Track changes of global encoding
@@ -528,12 +683,14 @@ proc GlobalEncoding {ref args} {
 
 proc Reconfigure {name args} {
 	variable icq
-	$icq configure -$name [set network::$name]
+	if {[catch { $icq configure -$name [set network::$name] } r]} {
+		Event Log error "Cant configure $name: $r"
+	}
 }
 
 proc CredChanged {what name args} {
 	upvar 1 $name val
-	Event Log 7 "$what changed to $val"
+	Event Log debug "$what changed to $val"
 	variable recreate 1
 	# Cancel registration of new UIN
 	Event ICQ:CancelRegistration
@@ -559,6 +716,9 @@ handler ConfigLoaded ConfigLoaded {args} {
 		trace variable $x w [nc Reconfigure [namespace tail $x]]
 	}
 
+	MonitorDCServer
+	trace variable direct::enable w [nc MonitorDCServer]
+
 	set icq [CreateConnection]
 	Event SetStatus $status 
 }
@@ -573,7 +733,11 @@ handler Send Send {type uid message} {
 		}
 		return -code continue
 	}
-	$icq send $type [lindex [split $uid :] 2] $message
+	set refdc [ref $uid](dc)
+	if {[lsearch {furtureauth authorization authrequest} $type]==-1 &&
+	    [info exists $refdc] && [[set $refdc] state]=="established"} {
+		[set $refdc] send $type $message
+	} else { $icq send $type [lindex [split $uid :] 2] $message }
 } 0.90
 
 handler Send SendSMS {type uid message} {
@@ -606,28 +770,47 @@ handler ChangeICQPassword ChangeICQPassword {password ref} {
 handler InfoRequest InfoRequest {uid ref} {
 	variable icq
 	if {![string match Contact:ICQ:* $uid]} return
-	$icq info [lindex [split $uid :] 2] $ref
+	set uin [lindex [split $uid :] 2]
+	if {[string is digit $uin]} { $icq info $uin $ref }
 }
 
-handler Incoming AddContact {type sender time message args} {
-	NewContact $sender
-} 0.01
-
-handler AddICQContact AddICQContact {uin groups alias} {
-	NewContact Contact:ICQ:$uin
+proc onOnline {} {
+	RequestInfo
 }
 
-proc NewContact {uid} {
-	if {[info exists [ref $uid](Groups)] ||
-	    ![string match Contact:ICQ:* $uid]} return
-	puts "adding $uid"
-	set uin [lindex [split $uid :] end]
-	new $uid [list Groups other Alias $uin]
+variable request_info [list]
+handler New:Contact:ICQ:* RequestAlias {uid} {
+	if {[info exists [ref $uid](Alias)]} return
+	variable request_info
+	lappend request_info $uid
+	if {[set [ref Me](Status)]!="offline"} { RequestInfo }
+} 0.06
+
+array set need_info [list]
+variable refcnt 100
+# Send info request on each item in need_info list
+proc RequestInfo {} {
+	variable request_info
+	if {![llength $request_info]} return
+	variable refcnt
+	variable need_info
 	variable icq
-	$icq contacts all add $uin
-	set ref [expr [clock seconds] & 0xff]
-	hook Info [nc NewContactInfo Contact:ICQ:$uin $ref]
-	$icq info $uin $ref
+	set uid [lindex $request_info 0]
+	$icq info [lindex [split $uid :] end] $refcnt
+	set need_info($refcnt) $uid
+	if {[incr refcnt]>250} { set refcnt 100 }
+	set request_info [lrange $request_info 1 end]
+	after 300 [nc RequestInfo]
+}
+
+handler Info Nick2Alias {ref info} {
+	variable need_info
+	if {![info exists need_info($ref)]} return
+	set uid $need_info($ref)
+	unset need_info($ref)
+	array set Info $info
+	if {![info exists Info(Nick)]} return
+	set [ref $uid](Alias) $Info(Nick)
 }
 
 # Sync icq alias when global alias changes
@@ -652,3 +835,18 @@ proc NewContactInfo {uid myref ref info} {
 	unhook Info [nc NewContactInfo $uid $ref]
 }
 
+handler ChatOpen PeerOpen {uid args} {
+	if {!$direct::enable} return
+	set ref [ref $uid]
+	if {[info exists ${ref}(dc)]} { 
+		set ${ref}(_is_open) 1
+		[set ${ref}(dc)] connect 
+	}
+}
+
+handler ChatClose PeerClose {uid args} {
+	set ref [ref $uid]
+	if {[info exists ${ref}(_is_open)]} { unset ${ref}(_is_open) }
+	if {[info exists ${ref}(dc)]} { [set ${ref}(dc)] disconnect }
+}
+
