@@ -20,6 +20,8 @@ option add *quick.Checkbutton.selectColor "" widgetDefault
 option add *quick.Checkbutton.indicatorOn 0
 option add *status.compound none widgetDefault
 option add *quickPanel true widgetDefault
+option add *rosterContactColor black widgetDefault
+option add *localContactColor blue widgetDefault
 
 namespace eval meta {
 	set author "Ihar Viarheichyk <iverg@mail.ru>"
@@ -30,6 +32,10 @@ namespace eval meta {
 		type variant default Plain property Global:UI:Tree|Grouping
 		description "Grouping rule" save exit menu {Interface Grouping}
 	}
+	array set quick_search {
+		type {boolean cache} default 0 menu {Search {In Contact List}}
+		weight .099
+	}
 }
 
 namespace eval hide {
@@ -49,7 +55,16 @@ namespace eval hide {
 	}
 }
 
-namespace eval filter {}
+# Visibility filters
+namespace eval filter {
+	# Contacts with pending messages are always visible
+	proc _pending {uid} {
+		if {[info exists [ref $uid](Pending)]} {
+			return -code break
+		} else { return 1}
+	}
+}
+
 namespace eval [ref Group:common]::meta {
 	array set Open	{type boolean default 0 save exit}
 }
@@ -89,14 +104,14 @@ proc MainWindow {} {
 	grid [checkbutton .quick.offline -image img:offline\
 		-variable [namespace current]::hide::offline]\
 		-row 0 -column 1 -padx 2 -ipadx 2 -ipady 2 -sticky w
-	tooltip .quick.offline $hide::meta::offline(description) <<Hide-Offline>>
+	ui::tooltip .quick.offline $hide::meta::offline(description) <<Hide-Offline>>
 	grid [checkbutton .quick.empty -image img:group\
 		-variable [namespace current]::hide::empty]\
 		-row 0 -column 2 -padx 2 -ipadx 2 -ipady 2 -sticky w
-	tooltip .quick.empty $hide::meta::empty(description) <<Hide-Empty>>
+	ui::tooltip .quick.empty $hide::meta::empty(description) <<Hide-Empty>>
 	grid [checkbutton .quick.search -text [mc Search] -variable $searchVar]\
 		-row 0 -column 10 -padx 2 -ipadx 2 -ipady 2 -sticky e
-	tooltip .quick.search "Quick search" <<QuickSearch>>
+	ui::tooltip .quick.search "Search in contact list" <<QuickSearch>>
 	trace variable $searchVar w [nc QuickSearch .qsearch .tr.c]
 	set sw [ScrolledWindow .sw -relief flat -borderwidth 0]
 	set tree [Tree .tr -deltay [DeltaY]\
@@ -107,6 +122,9 @@ proc MainWindow {} {
 		$tree bind${id} <Double-1> [nc DoubleClick]
 		$tree bind${id} <<ContextCurrent>> [nc PopupMenu %X %Y]
 	}
+	foreach x {roster local} { 
+		variable _${x} [option get .tr ${x}ContactColor ContactColor] 
+	}
 	bind .tr.c <<ContextSelected>> [nc PopupSelected %X %Y]
 
 	set menu1 [menubutton .main -text [mc "Menu"]\
@@ -116,8 +134,8 @@ proc MainWindow {} {
 		-menu .status.m -image img:offline]
 	set search [FastSearch .qsearch .tr $searchVar]
 	grid $sw -columnspan 2 -row 3 -sticky news
-	grid $menu2 -column 0 -row 5 -sticky w 
-	grid $menu1 -column 1 -row 5 -sticky w
+	grid $menu2 -column 0 -row 5 -sticky nws -padx 2 
+	grid $menu1 -column 1 -row 5 -sticky nes -padx 2
 	grid columnconfigure . 1 -weight 1
 	grid rowconfigure . 3 -weight 1
 
@@ -206,14 +224,8 @@ proc DoubleClick {node} {
 proc CreateMenus {} {
 	menu .status.m -type normal -tearoff no
 	foreach status {online away dnd occ ffc na invisible offline} {
-		if {$::tk_version >= 8.4} {
-			.status.m add command -image img:$status\
-				-label [mc $status] -compound left\
-				-command "Event SetStatus $status"
-		} else {	
-			.status.m add command -image img:$status\
-				-command "Event SetStatus $status"
-		}
+		.status.m add command -image img:$status -label [mc $status]\
+			-command "Event SetStatus $status"
 		set [ref Group:status:$status](Alias) [mc $status]
 	}
 }
@@ -233,6 +245,17 @@ proc ItemModify {uid text} {
 # --- Handle moving items between groups
 proc DropCmd {dest src type op datatype data} {
 	# Don't accept unknown types and don'w allow loops
+	if {$type=="widget" && $dest==".tr" ||
+	    [lindex $type 0]=="position" && [lindex $type 1]=="root"} {
+		set uid [UidByNode $data]
+		set parent [UidByNode [NodeParent $data]]
+		if {[string match Group:common:* $parent]} {
+			set parent [lindex [split $parent :] end]
+			set [ref $uid](Groups) [grep x [get $uid Groups {}]\
+				{$x!=$parent}]
+		}
+		return
+	}
 	if {![string match node* $type]||\
 	     [string match ${data}* [lindex $type 1]]} return
 	set dest [UidByNode [lindex $type 1]]
@@ -260,6 +283,10 @@ handler {*|Delete} Delete {id} {
 	if {[set pos [lsearch -exact $mapping($uid) $CurrentNode]]>=0} {
 		set lst [lreplace $mapping($uid) $pos $pos]
 		if {![llength $lst]} { 
+			set name [get $uid Alias]
+			if {[tk_messageBox -type yesno -title "Removing $name"\
+				-message "Remove $name from contact list?"]
+					!="yes"} return
 			unset [ref $uid](Groups)
 			after idle [list unset [ref $uid]]
 		} else {
@@ -348,6 +375,7 @@ proc onAlias {uid args} {
 
 proc onStatus {uid args} {
 	redraw $uid -image -helptext
+	foreach x [get $uid Groups] {delayed [nc redraw Group:common:$x -text]}
 	#if {[string is true $hide::offline] } { Mapping $uid }
 	Mapping $uid 
 	if {$sort::contacts=="Status"} { DoSort $uid }
@@ -368,16 +396,38 @@ proc NoPending {uid args} {
 	unhook Flashing [nc FlashIcon $uid]
 	FlashIcon $uid 1
 	Flash less
+	Mapping $uid
 	trace variable [ref $uid](Pending) w [nc DoPending $uid]
+	# Check if there other invisble contacts with pending messages
+	delayed [nc SeeNext]
 }
 
 proc DoPending {uid args} {
+	variable mapping
+	if {![info exists mapping($uid)]} { Mapping $uid }
+	See $uid
 	hook Flashing [nc FlashIcon $uid]
 	Flash more
 	trace vdelete [ref $uid](Pending) w [nc DoPending $uid]
 	trace variable [ref $uid](Pending) u [nc NoPending $uid]
 }
 
+# Find contact havind pending message and make it viewable in the tree
+proc SeeNext {} {
+	set uid [lindex [select Contact {[info exists Pending]}] 0]
+	if {$uid!=""} { See $uid }
+}
+
+# Resisively open parent groups and make any of object nodes visible
+proc See {uid} {
+	variable mapping
+	if {[info exists mapping($uid)]} {
+		foreach x $mapping($uid) { if {[.tr visible $x]} break }
+		if {![.tr visible $x]} { RecOpen .tr $x}
+		.tr see $x
+	}
+}
+
 # Namespace containing different grouping policies
 namespace eval Grouping {
 	# No grouping (default)
@@ -387,8 +437,8 @@ namespace eval Grouping {
 	# Use grous
 	proc Groups {uid} {
 		set ref [ref $uid]
-		if { ![info exists ${ref}(Groups)] } { return {} }
-		set g [set ${ref}(Groups)]
+		if { ![info exists ${ref}(Groups)] } { return "" }
+		set g [set ${ref}(Groups)] 
 		expr {[llength $g]?[map x $g {set _ Group:common:$x}]:{root}}
 	}
 	# Group by status
@@ -482,9 +532,9 @@ proc RemoveNode {node} {
 	}
 }
 
-proc UidByNode {node} { lindex [split $node /] end }
+proc UidByNode {node} { lindex [split $node "\0"] end }
 
-proc NodeParent {node} {string range $node 0 [expr [string last / $node]-1]}
+proc NodeParent {node} {string range $node 0 [expr [string last "\0" $node]-1]}
 
 proc redraw {uid args} {
 	variable mapping
@@ -494,18 +544,31 @@ proc redraw {uid args} {
 	}
 }
 
-proc node-text {uid} { get $uid Alias }
+proc node-text {uid} { 
+	if {[string match Group:common:* $uid]} {
+		set text [get $uid Alias]
+		foreach x {all online} { set $x 0 }
+		set g [lindex [split $uid :] end]
+		foreach x [select Contact "\[info exists Groups\] && \[lsearch \$Groups $g\]!=-1"] {
+			incr all
+			if {[get $x Status offline]!="offline"} { incr online }
+		}
+		append text " - $online/$all"
+	} else { get $uid Alias }
+}
 
 proc node-image {uid} {
 	set img group
-	if {[string match Contact:* $uid]} { set img [get $uid Status offline] }
+	if {[string match Contact:* $uid]} { 
+		set img [get $uid Status offline] 
+	}
 	if {[nil? $img]} { set img offline }
 	return img:$img
 }
 
 proc node-fill {uid} {
 	upvar #0 [ref $uid](rid) rid
-	expr {[info exists rid]?"black":"blue"}
+	set [namespace current]::[expr {[info exists rid]?"_roster":"_local"}]
 }
 
 proc node-open {uid} { 
@@ -526,21 +589,23 @@ if [package vsatisfies [package present BWidget] 1.6] {
 			append t "\n$client, ICQv${ver}\n"\
 				[expr ${unicode}?"Unicode":"Encoding: [get $uid encoding [set [pref {icq encoding}]]]"]
 		}
+		if {[info exists info(client:name)]} {
+			append t "\n$info(client:name)" 
+		}
 		if {[info exists info(Status)]} {
 			append t "\n[mc Status]: [mc $info(Status)]"
-		}	
+		}
+		if {[info exists info(description)]} {
+			append t "($info(description))"
+		}
+		if {[info exists info(IP)]} { append t "\nIP: $info(IP)" }
+		if {[info exists info(dc)] && [info exists info(LocalIP)]} {
+			append t "\nDC IP: $info(LocalIP)"
+		}
 		set t
 	}
-	proc ::tooltip {widget text {hotkey ""}} {
-		set text [mc $text]
-		if {$hotkey!=""} { append text "\n" [mc Hot-key] ": "\
-			[join [map x [event info $hotkey]\
-				{string map {Key- ""} $x}] ", "]}
-		DynamicHelp::register $widget balloon $text
-	}
 } else {
 	proc node-helptext {args} { return -code continue }
-	proc ::tooltip {args} {}
 }
 
 # Sort namespace includes all available sorting methods
@@ -593,6 +658,15 @@ namespace eval sort {
 	#Group-specific sorting
 	namespace eval groups {
 		namespace import [namespace parent]::*
+
+		if 0 {
+		proc Unsorted {s1 s2} {
+			puts "compare $s1 and $s2"
+			set w1 [get $s1 weight 0.5]
+			set w2 [get $s2 weight 0.5]
+			expr {($w1<$w2)?-1:(($w1==$s2)?0:1)}
+		}
+		}
 	}
 }
 
@@ -640,7 +714,18 @@ foreach x [info commands node-*] {
 }
 
 #Handle tree nodes mapping changes
-proc Mapping {uid args} { MappingChanged $uid [mapper $uid] }
+proc Mapping {uid args} { 
+	MappingChanged $uid [mapper $uid] 
+	if {[string match Group:common:* $uid]} {
+		set g [lindex [split $uid :] end]
+		foreach x [select Contact "\[lsearch \$Groups $g\]!=-1"] {
+			MappingChanged $x [mapper $x] 
+		}
+		redraw $uid -text
+	} elseif {[string match Contact:* $uid]} {
+		foreach x [get $uid Groups] { delayed [nc redraw Group:common:$x -text] }
+	}
+}
 
 proc MappingChanged {uid new} {
 	variable mapping
@@ -662,7 +747,7 @@ proc mapper {uid} {
 	set lst [list]
 	foreach x [grouper $uid] {
 		set m [expr {($x=="root")?"root":[mapper $x]}]
-		foreach y $m { lappend lst $y/$uid }
+		foreach y $m { lappend lst "$y\0$uid" }
 	}
 	set lst
 }
@@ -684,8 +769,8 @@ proc trigger {name} { set $name [expr [string is true [set $name]]^1] }
 
 proc UpdateStatus {args} {
 	upvar #0 [ref Me](Status) status
-	.status configure -image img:$status
-	tooltip .status [mc $status]
+	.status configure -image img:$status -text [mc $status]
+	ui::tooltip .status [mc $status]
 	WindowTitle
 }
 proc WindowTitle {args} {
@@ -693,6 +778,7 @@ proc WindowTitle {args} {
 }
 
 resource icons {group away dnd ffc invisible message na occ offline online url}
+
 MainWindow
 CreateMenus
 bind . <<Hide-Offline>> [nc trigger hide::offline]
