@@ -89,7 +89,7 @@ proc adjust {id} {
 
 proc connect {id server port script_ok script_error log} {
 	upvar #0 $id data 
-	eval $log 3 "{Connecting to proxy $data(-server):$data(-port)}"
+	eval $log info "{Connecting to proxy $data(-server):$data(-port)}"
 	if {[catch {
 		set sock [socket -async $data(-server) $data(-port)]
 	    } v]} { 
