@@ -1,20 +1,24 @@
-package provide alicq 0.8.8
+package provide alicq 0.8.9
 
 package require hooks 3.1
 package require helper
 
-set compatible 1
+set compatible 0
 
 # Core modules, provided by alicq itself.
 namespace eval modules {
 	# Global actions and properties
 	namespace eval my::meta {
+		array set add:group {type action menu {Add Group}
+			script { Event AddItem Group:common }
+		}
 		array set exit {type action menu Exit weight 0.9}
 	}
 }
 # Proxy servers management
 namespace eval modules::Proxy {
 	namespace eval meta {
+		set name Proxies
 		set description "Alicq core - proxies"
 		set author "Ihar Viarheichyk <iverg@mail.ru>"
 		set configuration-objects Proxy
@@ -39,6 +43,53 @@ namespace eval modules::Proxy {
 	}
 }
 
+namespace eval modules::log {
+	array set mapping {error 0 warning 1 notice 3 info 5 debug 10}
+	namespace eval meta {
+		set name Logging
+		set description "Alicq core - event logging"
+		set author "Ihar Viarheichyk <iverg@mail.ru>"
+		array set level { type variant save change }
+		set level(values) [array names [namespace parent]::mapping]
+	}
+	set level error
+	variable fd stderr
+
+	proc loglevel {str} {
+		variable mapping
+		if {![string is integer $str]} {
+			if {[info exists mapping($str)]} {
+				set str $mapping($str)
+			} else { set str 5 }
+		}
+		set str
+	}
+
+	proc init {name} {
+		variable fd
+		if {[catch {
+			if {[file exists $name]} {file delete $name}
+			set fd [open $name a+]
+			fconfigure $fd -buffering none
+		} reason]} {
+			set fd stderr
+			Event Log error "Can not open $name, using stderr: $reason"
+		}
+		Event Log 0 "Alicq started"
+	}
+
+	handler Log Log {tags str} {
+		variable level
+		variable fd
+		set l [lindex $tags 0]
+		if {[loglevel $l] > [loglevel $level]} return
+		if {[llength $tags]} { set tags "\[$tags\]" }
+		puts $fd [format "%s: %s %s" [clock format [clock seconds]\
+				-format {%x %X}] $tags $str]
+	}
+}
+interp alias {} log {} moduleconfig log
+
 # Namespace 'objects' contains all objects which should be managed with 
 # alicq helper procedures.
 namespace eval objects {
@@ -58,7 +109,7 @@ namespace eval objects {
 			array set Rename {type action menu Rename weight .98}
 			array set Delete {type action menu Delete weight .99}
 			array set Alias {default "" save change}
-			#array set Groups {type groups default {} save change menu Groups}
+			array set Groups {type groups default {} save change}
 		}
 	}
 	# Proxy servers
@@ -95,9 +146,6 @@ proc valid {metavar val} {
 	return 1
 }
 
-set LOGLEVEL 6
-set logfd stderr
-
 proc ref {obj} { return ::objects::[string map {: ::} $obj] }
 proc obj {ref} {join [lrange [split [string map {:: "\0"} $ref] "\0"] 2 end] :}
 
@@ -166,49 +214,20 @@ proc DefaultProperties {} {
 	property Group:*|Groups	[list -type list -default [nil] -save change]
 }
 
-proc DefaultLogging {} {
-	set ::LOGNAME [file join [file dirname $::RCFILE] alicq.log]
-	if {[catch {
-		if {[file exists $::LOGNAME]} {file delete $::LOGNAME}
-		set ::logfd [open $::LOGNAME a+]
-		fconfigure $::logfd -buffering none
-	} reason]} {
-		set ::logfd stderr
-		Event Log 0 "Can not open $::LOGNAME, using stderr: $reason"
-	}
-	Event Log 0 "Alicq started"
+# Contact command was extended to accept contacts of different messaging 
+# systems. Old-style ICQ command can be distingushed by 1st argument, which is 
+# numerical UIN in case of ICQ, and textual contact type otherwise.
+proc Contact {args} {
+	set type [expr {[string is digit [lindex $args 0]]?"old":"new"}]
+	eval Contact:$type $args
 }
 
-# Default resources and keybindings common for all modules
-proc DefaultTheme {} {
-	option add *Entry.background gray90 widgetDefault
-	option add *Text.background gray90 widgetDefault
-	option add *borderWidth 1 widgetDefault
-	option add *Button*overRelief raised widgetDefault
-	option add *highlightThickness 0 startupFile
-	option add *compound left widgetDefault
-	
-	switch -- $::tcl_platform(platform) {
-		unix {
-			option add *font -*-fixed-medium-r-normal--13-*\
-				widgetDefault
-		}
-	}
-	
-	event add <<Close>> <Escape>
-	event add <<Accept>> <Return>
-}
-
-proc AddObject {type name} {
-	switch $type {
-		group {Group [clock seconds] $name}
-		contact {Event AddICQContact $name {other} $name}
-	}
-}
-
-proc Contact {uin alias groups {properties {}}} {
+proc Contact:old {uin alias groups {properties {}}} {
 	Setter Contact:ICQ:$uin $properties [list Alias $alias Groups $groups]
 }
+
+proc Contact:new {type id properties} { new Contact:$type:$id $properties }
+
 proc Group {gid alias {groups {}} {properties {}}} { 
 	Setter Group:common:$gid $properties [list Alias $alias Groups $groups]
 }
@@ -223,11 +242,19 @@ proc Setter {id plist list} {
 
 proc new {id list} {
 	variable compatible
-	if {$compatible>1} {
-	   trace variable [ref $id] w [list [namespace current]::onChange $id]
+	set ref [ref $id]
+	if {[info exists $ref]} {
+		set event 0
+	} else {
+		if {$compatible>1} {
+			trace variable $ref w\
+				[list [namespace current]::onChange $id]
+		}		
+		set event 1
 	}
-	array set [ref $id] $list
-	Event New:$id $id
+	if {[catch { array set $ref $list } reason]} {
+		Event Log error "Can no create object $id: $reason"
+	} elseif {$event} { Event New:$id $id }
 }
 
 # DEBUG: map contact groperty to variables
@@ -285,26 +312,36 @@ if $compatible {
 proc Save {file key newval} {
 	set changed 0
 	set attr 0600
-	set fd2 [open ${file}.new w]
-	if {![catch {set fd1 [open $file r]}]} {
-		if {$::tcl_platform(platform)=={unix}} {
-			set attr [file attributes $file -permissions]
-		}	
-		seek $fd1 0
-		while {[gets $fd1 ln]!=-1} {
-			if {[regexp $key $ln]} {
-			 	if {![nil? $newval]} {puts $fd2 $newval}
-			 	set changed 1
-			} else { puts $fd2 $ln }
+	if {[catch { set fd2 [open ${file}.new w] } reason] } { 
+		Event Log error "Can not create ${file}.new: $reason"
+		return 
+	}
+	if {[file exists $file]} {
+		if {[catch {
+			if {$::tcl_platform(platform)=={unix}} {
+				set attr [file attributes $file -permissions]
+			}	
+			set fd1 [open $file r]
+			while {[gets $fd1 ln]!=-1} {
+				if {[regexp $key $ln]} {
+					if {![nil? $newval]} {puts $fd2 $newval}
+					set changed 1
+				} else { puts $fd2 $ln }
+			}
+			close $fd1
+		} reason]} { 
+			Event Log error "Error when modifying $file: $reason"
+			return
 		}
-		close $fd1
-	}	
-	if {!$changed && ![nil? $newval]} {puts $fd2 $newval }
-	close $fd2
-	file rename -force ${file}.new $file
-	if {$::tcl_platform(platform)=={unix}} {
-		file attributes $file -permissions $attr
-	}	
+	}
+	if {[catch {
+		if {!$changed && ![nil? $newval]} {puts $fd2 $newval }
+		close $fd2
+		file rename -force ${file}.new $file
+		if {$::tcl_platform(platform)=={unix}} {
+			file attributes $file -permissions $attr
+		}
+	} reason]} { Event Log error "Can't update $file: $reason"}	
 }
 
 
@@ -339,7 +376,7 @@ proc cmdline {} {
 		set cmd option_$opt
 		if {[llength [info commands $cmd]]} {
 			if {[catch {eval $cmd $val} v]} {
-				Log 0 "Error while processing option $opt: $v"
+				Event Log error "Error while processing option $opt: $v"
 			}
 		}
 	}
@@ -462,10 +499,11 @@ if {[package vsatisfies [package present Tcl] 8.4]} {
 
 proc LoadConfig {name} {
 	set base [file dirname $name]
+	set pwd [pwd]
 	if {[catch [list cd $base] v]} {
 		puts "Can not cd into $base: $v"
 		return 0
-	}
+	} else { set name [file tail $name] }
 	rename unknown unknown.config
 	proc unknown {args} {
 		set res ""
@@ -473,9 +511,9 @@ proc LoadConfig {name} {
 			# Check object namespace
 			set uid [join [lrange $args 0 end-1] :]
 			set ns [namespace qualifiers [ref $uid]]
-			if {[nsexists $ns]} {
+			if {[nsexists $ns] && $ns!="::objects"} {
 				new $uid [lindex $args end]
-			} else { Log 0 $v }
+			} else { Event Log warning $v }
 		} 
 		set res
 	}
@@ -483,7 +521,7 @@ proc LoadConfig {name} {
 	rename unknown {}
 	rename unknown.config unknown
 	if {$res} {
-		Log 0 "Error while loading startup file $name: $v"
+		Event Log error "Error while loading startup file $name: $v"
 		return 0
 	} else {
 		Event ConfigLoaded
@@ -498,12 +536,12 @@ proc moduleconfig {ns args} {
 	set var [namespace qualifiers $var]
 	set name [var2meta $var]
 	if {![llength [info vars $name]]} { 
-		Event Log 2 "No configuration parameter [lrange $args 0 end-1] in $ns"
+		Event Log warning "No configuration parameter [lrange $args 0 end-1] in $ns"
 		return ""
 	}
 	set new [lindex $args end]
 	if {![valid $name $new]} {
-		Event Log 2 "$ns $parm: $new is not valid $meta(type)"
+		Event Log error "$ns $parm: $new is not valid $meta(type)"
 		return ""
 	}
 	set $var $new
@@ -525,7 +563,11 @@ proc check_modules {name} {
 }
 proc load_modules {name file} { 
 	set ns [file tail [file rootname $name]]
-	namespace eval ::modules::$ns [list source $file]
+	if {[catch { namespace eval ::modules::$ns [list source $file] } r]} {
+		Event Log error "Error loading module $name: $r\n$::errorInfo"
+		catch { namespace delete ::modules::$ns }
+		return
+	}
 	metainfo:update $ns
 	if {[info commands $ns]=="" && 
 	    [llength [grep x [metainfo:module ::modules::$ns] {
@@ -605,22 +647,20 @@ proc setDefault {default var name1 name2 op} {
 proc resource {class names} {
 	set checker check_$class
 	if {![llength [info commands check_$class]]} {set checker empty_check}
-	set toload [list]
-	foreach x $names { if {![$checker $x]} {lappend toload $x} }
-	if {![llength $toload]} return
 	if {[llength [info commands option]]} {
 		set preferred [option get . ${class}Set ResourceSet]
 	} else { set preferred "" }
 	set fdirs [resource_path $class $preferred]
-	foreach name $toload {
+	foreach name $names {
 		set found 0
+		if {[$checker $name]} continue
 		foreach dir $fdirs {
 			foreach x [glob -nocomplain -directory $dir\
 							$name $name.*] {
 				if {[catch {load_$class $name $x} v]} {
-					Event Log 1 "$class: Error loading $name from $x: $v"
+					Event Log [list error $class] "Error loading $name from $x: $v"
 				} else {
-					Event Log 8 "$class: $name loaded from $x"
+					Event Log [list info $class] "$name loaded from $x"
 					set found 1
 					break
 				}
@@ -644,18 +684,39 @@ proc resource_path {class preferred} {
 	set path
 }
 
-# Logging
-handler Log Log {level str} {
-	if {$level>$::LOGLEVEL} return
-	set ln "\[[clock format [clock seconds] -format {%x %X}]\]: $str"
-	if {[catch { puts $::logfd $ln } reason]} {
-		Event Error output $reason
+# If sender is not in our contact list, add him into 'others' group.
+handler Incoming AddToOthers {type uid args} {
+	if {[info exists [ref $uid](Groups)]} return
+	new $uid {}
+} 0.05
+
+# If contact has no alias when created, assign default alias.
+# Protocol modules can handle this event with priority less than 0.1 
+# to request alias from server (if possible).
+handler {New:Contact:* New:Group:common:*} DefaultAlias  {uid} {
+	upvar 0 [ref $uid](Alias) alias
+	if {[info exists alias]} return
+	set alias [lindex [split $uid :] end]
+} 0.1
+
+handler {New:Contact:* New:Group:common:*} DefaultGroup  {uid} {
+	upvar 0 [ref $uid](Groups) groups
+	if {[info exists groups]} {
+		foreach x $groups {
+			if {![info exists [ref Group:common:$x]]} {
+				new Group:common:$x [list Alias $x]
+			}
+		}
+		return
+	} else {
+		if {[string match Contact:* $uid]} {
+			set groups other
+		} else { set groups "" }
 	}
-	Event ViewLog $ln
-} 
+} 0.1
 
 handler Error Error {id {description "Unknown error"}} {
-	Event Log 0 "ERROR $id \($description\)" 
+	Event Log error "ERROR $id \($description\)" 
 	list $id $description
 }
 
@@ -708,7 +769,7 @@ proc InitRC {} {
 			Error internal "Do not know where config file should be"
 			Finis
 		}
-		Log 0 "No startup file found, creating default"
+		Event Log warning "No startup file found, creating default"
 		set ::RCFILE [lindex $::rc 0]
 		set default [file join [file dirname [info script]] default.rc]
 		CopyConfig $default $::RCFILE
@@ -720,25 +781,25 @@ proc alicq_init {} {
 	# Detect available data paths
 	set ::DATA [list]
 	foreach x [list [file dirname $::RCFILE] [file join $mydir data]\
-			[file join [file dirname $mydir] share alicq]] {
+			[file join $mydir .. share alicq]] {
 		if {[file isdirectory $x]} { lappend ::DATA $x }
 	}
 	# Internatinalization
 	package require msgcat
-	namespace import msgcat::*
+	namespace import ::msgcat::*
 	foreach x $::DATA {
 		if {[msgcat::mcload [file join $x messages]]} break 
 	}
 	LoadConfig $::RCFILE
 }
 
-if {![catch {package present Tk}]} DefaultTheme
 # Parse command-line options
 cmdline
 # Check for default rc file
 InitRC
 # Default loggind
-DefaultLogging
+modules::log::init ${RCFILE}.log
 
 DefaultProperties
 
+
