@@ -43,7 +43,19 @@ proc lcompare {old new} {
 	list $comm $innew $inold
 }
 
-proc delayed {proc args} { foreach x {cancel idle} {after $x $proc $args} }
+# Detect if one of the programs exists in a path
+proc External {list} {
+	set path [split $::env(PATH) :]
+	foreach x $list {
+		foreach dir $path {
+			if {[file executable [file join $dir $x]]} { return $x }
+		}
+	}
+	return ""
+}
+
+#proc delayed {proc args} { foreach x {cancel idle} {after $x $proc $args} }
+proc delayed {args} { foreach x {cancel idle} {after $x $args} }
 proc nc {args} { uplevel 1 [list namespace code $args] }
 proc vars {args} { foreach x $args {uplevel 1 variable $x}}
 
